//
// UNROM512 (Mapper30) implementation.
// Based on Krikzz's mapper_2.v from http://krikzz.com/pub/support/everdrive-n8/development/fpga-mapper-sample.zip
//
module map_30
(map_out, cpu_dat, cpu_addr, cpu_ce, cpu_rw, m2,
ppu_dat, ppu_addr, ppu_oe, ppu_we, 
clk, map_cfg, rst, map_dout, map_oe, prg_mask);

	output [48:0]map_out;

	//cpu bus
	input [7:0]cpu_dat;
	input [14:0]cpu_addr;
	input cpu_ce, cpu_rw, m2;
	
	//ppu bus
	input [7:0]ppu_dat;
	input [13:0]ppu_addr;
	input ppu_oe, ppu_we;
	
	input clk, rst;
	input [7:0]map_cfg;
	output [7:0]map_dout;
	output map_oe;
	input [4:0]prg_mask;
	
	//mapper outputs
	reg  ciram_a10, ciram_ce;
	wire [19:0]prg_addr;
	wire [19:0]chr_addr;
	wire prg_we;
	wire chr_we;
	wire irq;
	wire rom_ce;
	wire ram_ce;
	reg  chr_ce;
	
	wire snd = 1'bzz;
	assign map_out[48:0] = {snd, chr_ce, ram_ce, rom_ce, irq, chr_we, prg_we, ciram_ce, ciram_a10, chr_addr[19:0], prg_addr[19:0]};
	//*************************************************************
	wire ram_area = cpu_addr[14:13] == 2'b11 & cpu_ce;
	wire flash_write = (write_state == 2'b11) & !cpu_ce & !cpu_addr[14];
	wire id_read = id_state & (!cpu_ce & !cpu_addr[14]) & (prg_addr[13:1] == 0) & cfg_battery;
	
	assign prg_we = !cpu_rw & flash_write;
	assign chr_we = map_cfg[7] ? !ppu_we & ciram_ce : 0;//if map_cfg[7] == 1 means that we don't have CHR rom, only CHR ram
	assign ram_ce = 0;
	assign rom_ce = !cpu_ce & !map_oe;
	
	assign mirroring_is_v = map_cfg[0];
	assign mirroring_is_4 = map_cfg[3];
	
	assign prg_addr[12:0] = cpu_addr[12:0];
	
	//Game should not write to upper 64K of battery ram, it may damage os code. 
	//So we pull upper lines of address to zero if not a rom area (cpu_ce == 1)
	assign prg_addr[13] = cpu_ce ? 0 : cpu_addr[13];
	assign prg_addr[18:14] = cpu_ce ? 0 : !cpu_addr[14] ? prg_bank[4:0] : 5'b11111;

	assign chr_addr[12:0] = ppu_addr[12:0];
	assign chr_addr[14:13] = chr_addr_14_13;
	
	reg [4:0]prg_bank;
	reg [1:0]chr_bank;
	reg nametable;
	reg [1:0]chr_addr_14_13;
	
	reg [1:0] write_state;
	reg id_state;
	
	wire [18:14] cfg_prgmask = prg_mask[4:0];
	
	// Software identification
	reg [7:0] manufacturer_id = 8'hED;  // ...Everdrive Flash Cart
	// For chip id, put a version number in top 4 bits
	wire [3:0] chip_id_version = 3;
	wire [3:0] chip_id_size = cfg_prgmask[18] ? 7 : // 512kB PRG
	                          cfg_prgmask[17] ? 6 : // 256kB PRG
                              cfg_prgmask[16] ? 5 : // 128kB PRG
                              cfg_prgmask[15] ? 4 : //  64kB PRG
                              cfg_prgmask[14] ? 3 : //  32kB PRG
                                                2;  //  16kB PRG
	wire [7:0] device_id = { chip_id_version, chip_id_size };

	assign map_dout = (prg_addr[0] ? device_id : manufacturer_id);
	assign map_oe = id_read;
	
	// Battery bit set:   write mapper at $C000-$FFFF
	// Battery bit clear: write mapper at $8000-$FFFF
	wire cfg_battery = map_cfg[1];
	wire mapper_reg = cfg_battery ? (!cpu_ce & cpu_addr[14]) : !cpu_ce;
	
	always @(negedge m2)
	begin
		if(!cpu_ce & !cpu_rw)
		begin
			if(mapper_reg) begin
				prg_bank[4:0] <= cpu_dat[4:0];
				chr_bank[1:0] <= cpu_dat[6:5];
				nametable <= cpu_dat[7];
				write_state <= write_state;
			end
			else begin
				// Advance state machine for byte-program sequence
				case(write_state)
					2'b00:begin
						write_state <= (prg_addr[14:0] == 15'h5555) & (cpu_dat == 8'hAA) ? 2'b01 : 2'b00;
					end
					2'b01:begin
						write_state <= (prg_addr[14:0] == 15'h2AAA) & (cpu_dat == 8'h55) ? 2'b10 : 2'b00;
					end
					2'b10:begin
						write_state <= (prg_addr[14:0] == 15'h5555) & (cpu_dat == 8'hA0) ? 2'b11 : 2'b00;
					end
					default:begin
						// Always go back to state 0 after writing
						// (writing handled asynchronously through flash_write signal)
						write_state <= 2'b00;
					end
				endcase
				// State machine for software identify sequence
				// This has the same sequence as the byte-program up until the last byte,
				// so we can just piggy-back on that one.
				if (!id_state) begin
					id_state <= (write_state == 2'b10) & (prg_addr[14:0] == 15'h5555) & (cpu_dat == 8'h90) ? 1 : id_state;
				end
				else begin
					id_state <= (cpu_dat == 8'hF0) ? 0 : id_state;
				end
			end
		end
		
	end
	
	always @(*)
	begin
		case({ mirroring_is_4, mirroring_is_v })
			2'b00:begin
				// horizontal mirroring
				ciram_a10 = ppu_addr[11];
				ciram_ce  = !ppu_addr[13];
				chr_ce = ciram_ce;
				chr_addr_14_13 = chr_bank[1:0];
			end
			2'b01:begin
				// vertical mirroring
				ciram_a10 = ppu_addr[10];
				ciram_ce  = !ppu_addr[13];
				chr_ce = ciram_ce;
				chr_addr_14_13 = chr_bank[1:0];
			end
			2'b11:begin
				// 4-screen mirroring
				ciram_a10 = nametable;
				ciram_ce = 1;
				chr_ce = 1;
				chr_addr_14_13 = ppu_addr[13] ? 2'b11 : chr_bank[1:0];
			end
			default:begin
				// one-screen mirroring
				ciram_a10 = nametable;
				ciram_ce = !ppu_addr[13];
				chr_ce = ciram_ce;
				chr_addr_14_13 = chr_bank[1:0];
			end
		endcase
	end

endmodule
